import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from active_bo_ros.PolicyModel.GaussianRBFModel import GaussianRBF
from active_bo_ros.AcquistionFunctions.ExpectedImprovement import ExpectedImprovement
from active_bo_ros.AcquistionFunctions.ProbabilityOfImprovement import ProbabilityOfImprovement
from active_bo_ros.AcquistionFunctions.ConfidenceBound import ConfidenceBound

from active_bo_ros.ReinforcementLearning.ContinuousMountainCar import Continuous_MountainCarEnv


class BayesianOptimization:
    def __init__(self, env, nr_steps, nr_init=3, acq='ei', nr_weights=6, policy_seed=None):
        self.env = env
        self.nr_init = nr_init
        self.acq = acq

        self.X = None
        self.Y = None
        self.GP = None

        self.episode = 0
        self.counter_array = np.empty((1, 1))
        self.best_reward = np.empty((1, 1))
        self.distance_penalty = 0

        self.nr_policy_weights = nr_weights
        self.nr_steps = nr_steps
        self.policy_seed = policy_seed
        self.lower_bound = -1.0
        self.upper_bound = 1.0

        self.policy_model = GaussianRBF(self.nr_policy_weights,
                                        self.nr_steps,
                                        self.policy_seed,
                                        self.lower_bound,
                                        self.upper_bound)

        self.eval_X = 100

    def reset_bo(self):
        self.counter_array = np.empty((1, 1))
        self.GP = None
        self.episode = 0
        self.best_reward = np.empty((1, 1))

    def runner(self, policy):
        done = False
        env_reward = 0.0
        step_count = 0

        for i in range(len(policy)):
            action = policy[i]
            output = self.env.step(action)

            env_reward += output[1]
            done = output[2]
            step_count += 1

            if done:
                self.counter_array = np.vstack((self.counter_array, step_count))
                break

            if not done and i == len(policy):
                distance = -(self.env.goal_position - output[0][0])
                env_reward += distance * self.distance_penalty
                self.counter_array = np.vstack((self.counter_array, step_count))

        self.env.reset()
        return env_reward, step_count

    def initialize(self):
        self.env.reset()
        self.reset_bo()

        self.X = np.zeros((self.nr_init, self.nr_policy_weights))
        self.Y = np.zeros((self.nr_init, 1))

        for i in range(self.nr_init):
            self.policy_model.random_policy()
            self.X[i, :] = self.policy_model.weights.T
            policy = self.policy_model.rollout()

            reward, step_count = self.runner(policy)

            self.Y[i] = reward

        self.GP = GaussianProcessRegressor(Matern(nu=1.5))
        self.GP.fit(self.X, self.Y)

    def next_observation(self):
        if self.acq == 'ei':
            x_next = ExpectedImprovement(self.GP,
                                         self.X,
                                         self.eval_X,
                                         self.nr_policy_weights,
                                         kappa=0,
                                         seed=self.policy_seed,
                                         lower=self.lower_bound,
                                         upper=self.upper_bound)

            return x_next

        elif self.acq == 'pi':
            x_next = ProbabilityOfImprovement(self.GP,
                                              self.X,
                                              self.eval_X,
                                              self.nr_policy_weights,
                                              kappa=0,
                                              seed=self.policy_seed,
                                              lower=self.lower_bound,
                                              upper=self.upper_bound)

            return x_next

        elif self.acq == 'cb':
            x_next = ConfidenceBound(self.GP,
                                     self.X,
                                     self.eval_X,
                                     self.nr_policy_weights,
                                     lam=2.576,
                                     seed=self.policy_seed,
                                     lower=self.lower_bound,
                                     upper=self.upper_bound)

            return x_next

        else:
            raise NotImplementedError

    def eval_new_observation(self, x_next):
        self.policy_model.weights = x_next
        policy = self.policy_model.rollout()

        reward, step_count = self.runner(policy)

        self.X = np.vstack((self.X, x_next))
        self.Y = np.vstack((self.Y, reward))

        self.GP.fit(self.X, self.Y)

        self.best_reward = np.vstack((self.best_reward, max(self.Y)))

        self.episode += 1
        return step_count



